
// OBJShape.java
// Andrew Davison, April 2007, ad@fivedots.coe.psu.ac.th

/* Display an OBJ Model using global arrays and other variables
   *PASTED* into the class before it's compiled. 

   The data is generated by ObjView, a Java 3D program that loads and
   displays an OBJ file, and outputs the globals into a textfile
   called examObj.txt. The contents of that file should be pasted
   in here.

   The model data allows a shape to utilize at most one
   material and/or texture.
 

   The data falls into 3 groups: model transformations, model
   coordinates, and material settings.

   // 1. model transformations

   private static final float xCenter, yCenter, zCenter;   
     // the center of the model; used to move the model to the origin

   private static final float scaleFactor;
     // used to scale the model so its longest dimension is 1 unit in length


   // 2. model coordinates

   private static final byte[] verts = { ... }
     // vertices for the model, scaled to be between -128-127, so each
        one will fit into a byte.
        The vertices will form triangle strips when rendered.

   private static final boolean hasNormals;   // true or false
   private static final byte[] normals = { ... }
     // normals for the model, scaled to be between -128-127, so each
        one will fit into a byte. This array may have no values, which
        means the model doesn't have normals. In that case, hasNormals
        will be false.

   private static final boolean hasTexture;   // true or false
   private static final float[] texCoords; = { ... }
     // tex coords for the model, in the form of floats between 0 and 1.0f.
        This array may have no values, which means the model doesn't use
        texturing. In that case, hasTexture will be false.

     // OpenGL ES can only deal with 2D tex coords, although it is 
        possibly for ObjView to output 3D and 4D coordinates. But I've never
        seen an OBJ model that uses them.

   private static final int[] strips = {...}
     // the number of vertices in each triangle strip making up
        the model

   There's no colour[] array since Java 3D's ObjFile class
   doesn't support colour coordinates in an OBJ file, 
   so there's no way that ObjView can generate them.


   // 3. material settings

   private static final String TEX_FNM = "/...";
      // the texture filename. This should be placed in the res/
         directory if you're using WTK.
         This string will always be the name of the OBJ file, but
         with a ".png" extension. If that's wrong, then you will 
         have to edit it.

   private static final float[] ambientMat, emissiveMat, diffuseMat, specularMat;
     // the RGBA values for the ambient, emissive, diffuse, and
        specular properties of the material. They will always have values.

   private static final float shininess;   // the shinness value (ranging from 0 to 128)
       // it will always have a value

   ------------------------------
   
   The object can be positioned and scaled, by supplying
   constructor arguments.
*/


import java.nio.*;
import javax.microedition.khronos.egl.*;
import javax.microedition.khronos.opengles.*;
import javax.microedition.lcdui.*;


public class OBJShape
{
  // ------------------ start model data --------------------------
  // PASTE in the contents of examObj.txt here



  // ------------------ end model data --------------------------



  private GL10 gl;

  // buffers for the model data arrays
  private ByteBuffer vertsBuf;   // vertices
  private ByteBuffer normsBuf;   // normal coords
  private FloatBuffer tcsBuf;    // tex coords (as floats)
  private ByteBuffer texBuf;     // RGB texture image

  private int imWidth, imHeight;   // dimensions of texture image
  private int texNames[];   // for the texture name

  private float xPos, yPos, zPos;  // position of model's center
  private float scale;


  public OBJShape(GL10 gl)
  {  this(gl, 0, 0, 0, 1);  }


  public OBJShape(GL10 gl, float x, float y, float z, float sc)
  // (x,y,z) is the model's position, and sc the scaling factor
  {
    this.gl = gl;
    xPos = x; yPos = y; zPos = z;
    scale = sc;

    // create vertices buffer
    vertsBuf = ByteBuffer.allocateDirect(verts.length);
    vertsBuf.put(verts).rewind();

    if (hasNormals) {   // create normals buffer
      normsBuf = ByteBuffer.allocateDirect(normals.length);
      normsBuf.put(normals).rewind();
    }
   
    if (hasTexture) {  // create texture data
      ByteBuffer bb = ByteBuffer.allocateDirect(texCoords.length * 4);
	  tcsBuf = bb.asFloatBuffer();
	  tcsBuf.put(texCoords).rewind();

      loadTexture(TEX_FNM);

      // generate a texture name
      texNames = new int[1];
      gl.glGenTextures(1, texNames, 0); 
    }
  }  // end of OBJShape()



  private void loadTexture(String fnm)
  // load the image stored in fnm into a byte buffer
  {
    try {
      Image image = Image.createImage(fnm);
      imWidth = image.getWidth();
      imHeight = image.getHeight();
      int pixels[] = new int[imWidth];   // for a row of pixels

      texBuf = ByteBuffer.allocateDirect(imWidth * imHeight * 3);

      /* Read the image in reverse row order to flip the image
         vertically in the buffer. This will order the texture 
         correctly for OpenGL's coordinate system, where (0,0) 
         is the lower left. */
      for (int i = imHeight-1; i >= 0; i--) {
        image.getRGB(pixels, 0, imWidth, 0, i, imWidth, 1);  // get a pixel row
        for (int j = 0; j < imWidth; j++) {
          // store each pixel in the row as three bytes
          texBuf.put((byte) (pixels[j] >> 16 & 0xff));  // red
          texBuf.put((byte) (pixels[j] >> 8 & 0xff));   // green
          texBuf.put((byte) (pixels[j] >> 0 & 0xff));   // blue
        }
      }
      texBuf.rewind();
    }
    catch (Exception e) 
    {  System.out.println("Error creating texture for " + fnm); }
  }  // end of loadTexture()



  public void draw()
  // draw the model
  {
    gl.glEnableClientState(GL10.GL_VERTEX_ARRAY);
    gl.glVertexPointer(3, GL10.GL_BYTE, 0, vertsBuf);  // use  verts

    if (hasNormals) {
      gl.glEnableClientState(GL10.GL_NORMAL_ARRAY);
      gl.glNormalPointer(GL10.GL_BYTE, 0, normsBuf);  // use normals
    }

    if (hasTexture) {
      gl.glEnable(GL10.GL_TEXTURE_2D);   // use texturing
      gl.glEnableClientState(GL10.GL_TEXTURE_COORD_ARRAY); 
      gl.glTexCoordPointer(2, GL10.GL_FLOAT, 0, tcsBuf);  // use tex coords
      setTexture();
    }

    setMaterial();

    gl.glPushMatrix();

      // move and scale the model using the user's settings
      gl.glTranslatef(xPos, yPos, zPos);   // move to the specified (x,y,z) pos
      if (scale != 1.0f)
        gl.glScalef(scale, scale, scale);  // uniform scaling

      // center and scale the model
      gl.glScalef(scaleFactor, scaleFactor, scaleFactor);  // uniform scaling
      gl.glTranslatef(-xCenter, -yCenter, -zCenter);   // move to the origin

      int pos = 0;
      int stripLen;
      for (int i = 0; i < strips.length; i++) {  // draw each strip of model
        gl.glDrawArrays(GL10.GL_TRIANGLE_STRIP, pos, strips[i]);  
                  // vertices, (normals,) (tex coords) for the strip
        pos += strips[i];
      }

    gl.glPopMatrix();

    gl.glDisableClientState(GL10.GL_VERTEX_ARRAY);
    gl.glDisableClientState(GL10.GL_NORMAL_ARRAY);
    gl.glDisableClientState(GL10.GL_TEXTURE_COORD_ARRAY); 

    gl.glDisable(GL10.GL_TEXTURE_2D);
  }  // end of draw()


  private void setTexture()
  {
    gl.glBindTexture(GL10.GL_TEXTURE_2D, texNames[0]);  // use the tex name

    // specify the texture for the currently bound tex name
    gl.glTexImage2D(GL10.GL_TEXTURE_2D, 0, GL10.GL_RGB, imWidth, imHeight, 0, 
                    GL10.GL_RGB, GL10.GL_UNSIGNED_BYTE, texBuf);

    // set the minification/magnification techniques
    gl.glTexParameterx(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MIN_FILTER, GL10.GL_LINEAR);
    gl.glTexParameterx(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MAG_FILTER, GL10.GL_LINEAR);
  } // end of setTexture()


  private void setMaterial()
  {
    gl.glMaterialfv(GL10.GL_FRONT_AND_BACK, GL10.GL_AMBIENT, ambientMat, 0);
    gl.glMaterialfv(GL10.GL_FRONT_AND_BACK, GL10.GL_DIFFUSE, diffuseMat, 0);    
    gl.glMaterialfv(GL10.GL_FRONT_AND_BACK, GL10.GL_SPECULAR, specularMat, 0);
    gl.glMaterialf(GL10.GL_FRONT_AND_BACK, GL10.GL_SHININESS, shininess);
    gl.glMaterialfv(GL10.GL_FRONT_AND_BACK, GL10.GL_EMISSION, emissiveMat, 0);
  }  // end of setMaterial()

}  // end of OBJShape class
