
// Floor.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, April 2007

/* The Floor is a square of length size, centered at the
   origin on the XZ plane. Wrapped over the top of it is
   the image loaded from florrFnm.

   The floor geometry is a triangle strip defining a square.

   This version of floor switches off lighting in its draw()
   method, so it won't be affected by the material settings
   for the model in OBJModel.
   This means that the floor doesn't need normals.
*/

import java.nio.*;
import javax.microedition.lcdui.*;
import javax.microedition.khronos.egl.*;
import javax.microedition.khronos.opengles.*;



public class Floor 
{
  private GL10 gl;

  private ByteBuffer vertsBuf;   // floor vertices
  private ByteBuffer tcsBuf;     // tex coords
  private ByteBuffer texBuf;     // RGB texture image
  private int imWidth, imHeight;   // dimensions of the texture image
  private int texNames[];   // for the texture name


  public Floor(GL10 gl, String floorFnm, int size) 
  { 
    this.gl = gl;

    createFloor(size);
    loadTexture(floorFnm);

    // generate a texture name
    texNames = new int[1];
    gl.glGenTextures(1, texNames, 0); 
  }  // end of Floor()


  private void createFloor(int size)
  // create the vertices and tex coords buffers for the floor of size-by-size
  {
    // create vertices buffer
    byte pt = (byte)(size/2);
    byte npt = (byte) (-size/2);   // negative pt
    // byte[] verts = {npt,0,pt,  pt,0,pt,  pt,0,npt,  npt,0,npt};
    byte[] verts = {npt,0,pt,  pt,0,pt,  npt,0,npt,  pt,0,npt};
    vertsBuf = ByteBuffer.allocateDirect(verts.length);
    vertsBuf.put(verts).rewind();

    // create texture coordinates buffer
    byte[] tcs = {0,0,  1,0,  0,1,  1,1};
    tcsBuf = ByteBuffer.allocateDirect(tcs.length);
    tcsBuf.put(tcs).rewind();
  } // end of createFloor()



  private void loadTexture(String fnm)
  // load the image stored in fnm into a byte buffer
  {
    try {
      Image image = Image.createImage(fnm);
      imWidth = image.getWidth();
      imHeight = image.getHeight();
      int pixels[] = new int[imWidth];   // for a row of pixels

      texBuf = ByteBuffer.allocateDirect(imWidth * imHeight * 3);

      /* Read the image in reverse row order to flip the image
         vertically in the buffer. This will order the texture 
         correctly for OpenGL's coordinate system, where (0,0) 
         is the lower left. */
      for (int i = imHeight-1; i >= 0; i--) {
        image.getRGB(pixels, 0, imWidth, 0, i, imWidth, 1);  // get a pixel row
        for (int j = 0; j < imWidth; j++) {
          // store each pixel in the row as three bytes
          texBuf.put((byte) (pixels[j] >> 16 & 0xff));  // red
          texBuf.put((byte) (pixels[j] >> 8 & 0xff));   // green
          texBuf.put((byte) (pixels[j] >> 0 & 0xff));   // blue
        }
      }
      texBuf.rewind();
    }
    catch (Exception e) 
    {  System.out.println("Error creating texture for " + fnm); }
  }  // end of loadTexture()



  public void draw()
  // called from RotBoxGameCanvas to draw the floor
  {
    gl.glDisable(GL10.GL_LIGHTING);    // disable lighting
    gl.glEnable(GL10.GL_TEXTURE_2D);   // use texturing

    /* enable arrays when rendering 
      (no normals in this version of Floor) */
    gl.glEnableClientState(GL10.GL_VERTEX_ARRAY);
    gl.glEnableClientState(GL10.GL_TEXTURE_COORD_ARRAY);

    gl.glVertexPointer(3, GL10.GL_BYTE, 0, vertsBuf);  // use floor verts
    gl.glTexCoordPointer(2, GL10.GL_BYTE, 0, tcsBuf);  // use tex coords
    setTexture();

    gl.glDrawArrays(GL10.GL_TRIANGLE_STRIP, 0, 4);

    // disable the arrays at the end of rendering
    gl.glDisableClientState(GL10.GL_VERTEX_ARRAY);
    gl.glDisableClientState(GL10.GL_TEXTURE_COORD_ARRAY); 

    gl.glDisable(GL10.GL_TEXTURE_2D);   // switch off texturing
    gl.glEnable(GL10.GL_LIGHTING);      // switch lighting back on
  }  // end of draw()


  private void setTexture()
  {
    gl.glBindTexture(GL10.GL_TEXTURE_2D, texNames[0]);  // use the tex name

    // specify the texture for the currently bound tex name
    gl.glTexImage2D(GL10.GL_TEXTURE_2D, 0, GL10.GL_RGB, imWidth, imHeight, 0, 
                    GL10.GL_RGB, GL10.GL_UNSIGNED_BYTE, texBuf);

    // set the minification/magnification techniques
    gl.glTexParameterx(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MIN_FILTER, GL10.GL_LINEAR);
    gl.glTexParameterx(GL10.GL_TEXTURE_2D, GL10.GL_TEXTURE_MAG_FILTER, GL10.GL_LINEAR);
  } // end of setTexture()


} // end of Floor Class
