
// KeyCamera.java
// Andrew Davison, ad@fivedots.coe.psu.ac.th, April 2007

/* Use key presses to move/rotate the camera.
   KeyCamera is an adapter for the Camera class

   The keys are grouped into 3 modes: TRANSLATE (for translations),
   ROTATE (for rotations), and FLOAT_GLIDE (for floating and gliding).
   They can be cycled through by the user pressing the <game-A> key.

   The keys used:
     * the arrow keys for translation and rotation;
     * <enter> to trigger a reset of the camera's position/orientation;

   System bug:
      WTK 2.5 has a problem with getKeyStates() when a key is
      held down for a while, and then released. It seems like 
      the state isn't completely cleared when the key is released, 
      causing the key press to be actied upon forever.
*/

import javax.microedition.lcdui.game.GameCanvas;
import javax.microedition.khronos.egl.*;
import javax.microedition.khronos.opengles.*;


public class KeyCamera
{
  // key modes
  private static final int TRANSLATE = 0;
  private static final int ROTATE = 1;
  private static final int FLOAT_GLIDE = 2;
  private static final int NUM_MODES = 3; // number of key modes

  // initial camera position
  private static final double X_POS = 0.0; 
  private static final double Y_POS = 1.5; 
  private static final double Z_POS = 5.0;

  // translation and rotation increments
  private static final double MOVE_INCR = 0.1;
  private static final double ANGLE_INCR = 5.0;   // degrees

  // navigation related
  private Camera camera;
  private boolean upKey, downKey, leftKey, rightKey, fireKey, gameAKey; // action keys
  private int keyMode;


  public KeyCamera(GL10 gl)
  {  camera = new Camera(gl, X_POS, Y_POS, Z_POS);  }


  // ------------------ update the camera state ---------------


  public void update(int keysState)
  /* use the currently pressed keys to update the camera's
     position and orientation */
  {
    readKeys(keysState);
    updateCamera();
  }


  private void readKeys(int keysState) 
  // record the key presses
  {
    // reset key booleans
    upKey = false;
    downKey = false;
    leftKey = false;
    rightKey = false;
    fireKey = false;
    gameAKey = false;

    // record key presses as booleans
    if ((keysState & GameCanvas.LEFT_PRESSED) != 0) 
      leftKey = true;
    if ((keysState & GameCanvas.RIGHT_PRESSED) != 0) 
      rightKey = true;
    if ((keysState & GameCanvas.UP_PRESSED) != 0) 
      upKey = true;
    if ((keysState & GameCanvas.DOWN_PRESSED) != 0) 
      downKey = true;
    if ((keysState & GameCanvas.FIRE_PRESSED) != 0) 
      fireKey = true;
    if ((keysState & GameCanvas.GAME_A_PRESSED) != 0) 
      gameAKey = true;
/*
    System.out.println("Keys state (U, D, L, R, F, GA) : (" + 
               upKey + ", " +  downKey + ", " +
               leftKey + ", " + rightKey + ", " + 
               fireKey + ", " + gameAKey + ")" );
*/
  } // end of readKeys()



  private void updateCamera()
  /* The three modes (TRANSLATE, ROTATE, and FLOAT_GLIDE) 
     cause the key presses to be interpreted differently.
  */
  {
    // cycle the key mode
    if (gameAKey)
      keyMode = (keyMode+1) % NUM_MODES;

    // reset the camera position/orientation
    if (fireKey)
      camera.reset();

    if (keyMode == TRANSLATE) { 
      if (upKey)
        camera.transFwd(MOVE_INCR);     // translate forward
      if (downKey)
        camera.transFwd(-MOVE_INCR);    // translate backward
      if (leftKey)
        camera.glideRight(-MOVE_INCR);  // move left parallel to XZ plane
      if (rightKey)
        camera.glideRight(MOVE_INCR);   // move right parallel to XZ plane
    }
    else if (keyMode == ROTATE) {
      if (upKey)
        camera.rotateX(ANGLE_INCR);    // rotate camera up around x-axis
      if (downKey)
        camera.rotateX(-ANGLE_INCR);   // rotate camera down
      if (leftKey)
        camera.rotateY(ANGLE_INCR);    // rotate camera left around y-axis
      if (rightKey)
        camera.rotateY(-ANGLE_INCR);   // rotate camera right
    }
    else if (keyMode == FLOAT_GLIDE) {
      if (upKey)
        camera.moveUp(MOVE_INCR);     // move camera up
      if (downKey)
        camera.moveUp(-MOVE_INCR);    // move camera down
      if (leftKey) 
        camera.glideFwd(MOVE_INCR);   // move forward parallel to XZ plane
      if (rightKey)
        camera.glideFwd(-MOVE_INCR);  // move backward parallel to XZ plane
    }
    else  // should not happen
      System.out.println("Unknown key mode");
  }  // end of updateCamera()



  public void position()
  // position the camera
  { camera.position();  }


  // --------------------- get camera state ------------------

  public String getMode()
  // return the camera mode
  { 
    if (keyMode == TRANSLATE) 
      return "Translate";
    else if (keyMode == ROTATE)
      return "Rotate";
    else if (keyMode == FLOAT_GLIDE)
      return "Float/glide";
    else
      return "Mode ??";
  }  // end of getMode()


  public String getPos()
  // return the current camera position
  {  return camera.getPos(); }

  public String getRots()
  // return the current camera rotations
  {  return camera.getRots();  }


}  // end of KeyCamera class